from __future__ import annotations

from typing import Optional, Any
import os
import h5py
from silx.gui import qt
from ewoksutils.task_utils import task_inputs


def generateInputs(
    newFlat: str,
    oldFlat: str,
    energy: float,
    pyfaiConfig: dict,
    pyfaiMethod: str,
    datasetFilename: str,
    scanNumber: int,
    monitorName: str,
    referenceCounts: int,
    detectorName: str,
    outputDirectory: str,
    sigmaClippingThreshold: Optional[float],
    asciiExport: bool,
    unit: str,
    tomoEnabled: bool = False,
    tomoScanEntryUrl: str = "",
    rotName: str = "",
    yName: str = "",
    tomoOutputFilename: str = "",
) -> dict:
    """
    Generate input parameters for the EWOKS workflow, including optional tomo export.
    """
    baseDirName = os.path.splitext(os.path.basename(datasetFilename))[0]
    outputUnitSuffix = unit.split("_")[0]
    outputFilePathH5 = os.path.join(outputDirectory, f"{baseDirName}.h5")

    externalOutputFilePathH5 = os.path.join(
        outputDirectory, f"{baseDirName}_{outputUnitSuffix}.h5"
    )
    nxProcessName = f"{detectorName}_{outputUnitSuffix}_integrate"
    nxMeasurementName = f"{detectorName}_{outputUnitSuffix}_integrated"

    outputAsciiFileTemplate = os.path.join(
        f"{baseDirName}_{scanNumber:04d}_{detectorName}_{outputUnitSuffix}_%04d.xye",
    )

    outputArchiveFilename = os.path.join(
        outputDirectory,
        "export",
        f"{baseDirName}_{scanNumber:04d}_{detectorName}_{outputUnitSuffix}.zip",
    )

    integrationOptions: dict[str, Any] = {
        "method": pyfaiMethod,
        "unit": unit,
    }
    if sigmaClippingThreshold is not None:
        integrationOptions["extra_options"] = {
            "thres": sigmaClippingThreshold,
            "max_iter": 10,
            "error_model": "azimuthal",
        }
        integrationOptions["integrator_name"] = "sigma_clip_ng"

    inputs = [
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "newflat",
            "value": newFlat,
        },
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "oldflat",
            "value": oldFlat,
        },
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "energy",
            "value": energy,
        },
        {
            "task_identifier": "PyFaiConfig",
            "name": "filename",
            "value": pyfaiConfig,
        },
        {
            "task_identifier": "PyFaiConfig",
            "name": "integration_options",
            "value": integrationOptions,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "filename",
            "value": datasetFilename,
        },
        {"task_identifier": "IntegrateBlissScan", "name": "scan", "value": scanNumber},
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "output_filename",
            "value": outputFilePathH5,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "monitor_name",
            "value": monitorName,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "reference",
            "value": referenceCounts,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "maximum_persistent_workers",
            "value": 1,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "retry_timeout",
            "value": 3600,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "detector_name",
            "value": detectorName,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "nxprocess_name",
            "value": nxProcessName,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "nxmeasurement_name",
            "value": nxMeasurementName,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "external_output_filename",
            "value": externalOutputFilePathH5,
        },
        {
            "task_identifier": "SaveNexusPatternsAsAscii",
            "name": "enabled",
            "value": asciiExport,
        },
        {
            "task_identifier": "SaveNexusPatternsAsAscii",
            "name": "output_filename_template",
            "value": outputAsciiFileTemplate,
        },
        {
            "task_identifier": "SaveNexusPatternsAsAscii",
            "name": "output_archive_filename",
            "value": outputArchiveFilename,
        },
        {
            "name": "overwrite",
            "value": True,
            "all": True,
        },
    ]

    inputs += task_inputs(
        task_identifier="SaveNexusPatternsAsId31TomoHdf5",
        inputs={
            "enabled": tomoEnabled,
            "scan_entry_url": tomoScanEntryUrl,
            "rot_name": rotName,
            "y_name": yName,
            "output_filename": tomoOutputFilename,
        },
    )

    return {
        "inputs": inputs,
        "convert_destination": os.path.join(
            outputDirectory,
            f"{baseDirName}_{scanNumber}_{detectorName}_{outputUnitSuffix}.json",
        ),
    }


def extractScanNumber(h5path: str) -> int:
    """
    Extracts the scan number from the h5path from a selected node.

    Example: '/2.1/measurement/p3" -> returns 2'
    """
    parts = h5path.split("/")
    if len(parts) > 1:
        scanNumberPart = parts[1].split(".")[0]
        try:
            return int(scanNumberPart)
        except ValueError:
            pass
    return -1


def generateUniquePath(basePath: str) -> str:
    """
    Generates a unique path by adding an incremantal suffix if necessary.

    Args:
        basePath: Base path (file or directory)

    Returns:
        unique path
    """
    isFile = os.path.isfile(basePath)
    dirName = os.path.dirname(basePath)
    baseName, ext = (
        os.path.splitext(os.path.basename(basePath)) if isFile else (basePath, "")
    )

    parts = baseName.rsplit("_", 1)
    if len(parts) == 2 and parts[1].isdigit():
        baseName, counter = parts[0], int(parts[1]) + 1
    else:
        counter = 1

    generatedName = os.path.join(dirName, f"{baseName}_{counter}{ext}")
    while os.path.exists(generatedName):
        counter += 1
        generatedName = os.path.join(dirName, f"{baseName}_{counter}{ext}")
    return generatedName


def getScanEnergy(h5Filename: str, scanNumber: int) -> Optional[float]:
    """
    Retrieves the energy for a specific scan in an HDF5 file.

    Args:
        h5Filename: Path to the HDF5 file.
        scanNumber: Scan number to retrieve energy for.

    Returns:
        Energy value for a given scan.
    """
    scanPath = f"{scanNumber}.1/instrument/positioners"

    with h5py.File(h5Filename, "r") as h5file:
        scan_group = h5file.get(scanPath)
        if not isinstance(scan_group, h5py.Group):
            return None
        energy_dset = scan_group.get("energy")
        if not isinstance(energy_dset, h5py.Dataset):
            return None
        return energy_dset[()]

    return None


class FilenameCompleterLineEdit(qt.QLineEdit):
    """
    Heritage from QLineEdit widget that provides autocompletion for file paths.

    This widget uses a QFileSystemModel to suggest file and directory paths,
    starting from the root of the filesystem.
    """

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent=parent, **kwargs)

        completer = qt.QCompleter()
        model = qt.QFileSystemModel(completer)
        model.setOption(qt.QFileSystemModel.Option.DontWatchForChanges, True)
        model.setRootPath("/")

        completer.setModel(model)
        completer.setCompletionRole(qt.QFileSystemModel.Roles.FileNameRole)
        self.setCompleter(completer)


def get_scan_url(workflow_parameters):
    datasetFilename = None
    scanNumber = None
    for item in workflow_parameters["inputs"]:
        if item.get("task_identifier") != "IntegrateBlissScan":
            continue
        if item["name"] == "filename":
            datasetFilename = item["value"]
        if item["name"] == "scan":
            scanNumber = item["value"]

    return datasetFilename, scanNumber



"""
Usage:

    python scripts/reprocess_xrpd.py --proposal ma5876 \
        --samples 02_cell1 03_cell2 \
        --pyfai-config /data/visitor/ma5876/id31/20231117/PROCESSED_DATA/calibration_good/20231117_json_good_bottom_beam_q_MM.json \
        --run reprocess_q_folder

    python scripts/reprocess_xrpd.py --proposal ma5876 \
        --samples 02_cell1 03_cell2 \
        --pyfai-config /data/visitor/ma5876/id31/20231117/PROCESSED_DATA/calibration_good/20231117_json_good_bottom_beam_2th_MM.json \
        --run reprocess_2th_folder

"""

import json
import os
import sys
from glob import glob
import argparse
import traceback

import h5py
from ewoks import execute_graph
from ewoksjob.client import submit
from ewoksutils.task_utils import task_inputs


from .utils import FLATFIELD_DEFAULT_DIR, NEWFLAT_FILENAME, OLDFLAT_FILENAME


WORKFLOW = "integrate_with_saving_with_flat.json"
WORKFLOW_LOAD_OPTIONS = {"root_module": "ewoksid31.workflows"}

DETECTOR = "p3"
BEACON_HOST = "id31:25000"
ENERGY_POSITIONER = "energy"


def get_parameters(
    dataset_filename,
    scan_number,
    detector_name,
    pyfai_config,
    pyfai_method,
    energy,
    monitor_name=None,
    reference_counts=None,
    run="",
    save_as_ascii=False,
):
    if run:
        output_filename = dataset_filename.replace("RAW_DATA", f"PROCESSED_DATA/{run}")
        convert_destination = dataset_filename.replace(
            "RAW_DATA", f"PROCESSED_DATA/{run}"
        )
    else:
        output_filename = dataset_filename.replace("RAW_DATA", "PROCESSED_DATA")
        convert_destination = dataset_filename.replace("RAW_DATA", "PROCESSED_DATA")

    output_filename_template = output_filename.replace(
        ".h5", f"_{scan_number}_{detector_name}_%04d.xye"
    )

    convert_destination = convert_destination.replace(
        ".h5", f"_{scan_number}_{detector_name}.json"
    )

    inputs = [
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "newflat",
            "value": os.path.join(FLATFIELD_DEFAULT_DIR, NEWFLAT_FILENAME),
        },
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "oldflat",
            "value": os.path.join(FLATFIELD_DEFAULT_DIR, OLDFLAT_FILENAME),
        },
        {"task_identifier": "FlatFieldFromEnergy", "name": "energy", "value": energy},
        {"task_identifier": "PyFaiConfig", "name": "filename", "value": pyfai_config},
        {
            "task_identifier": "PyFaiConfig",
            "name": "integration_options",
            "value": {"method": pyfai_method},
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "filename",
            "value": dataset_filename,
        },
        {"task_identifier": "IntegrateBlissScan", "name": "scan", "value": scan_number},
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "output_filename",
            "value": output_filename,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "monitor_name",
            "value": monitor_name,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "reference",
            "value": reference_counts,
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "maximum_persistent_workers",
            "value": 1,
        },
        {
            # Use a very long retry timeout to mitigate issues
            # with multiple workflows accessing the same files
            "task_identifier": "IntegrateBlissScan",
            "name": "retry_timeout",
            "value": 6 * 3600,  # 6h hours
        },
        {
            "task_identifier": "IntegrateBlissScan",
            "name": "detector_name",
            "value": detector_name,
        },
        {
            "task_identifier": "SaveNexusPatternsAsAscii",
            "name": "output_filename_template",
            "value": output_filename_template,
        },
        {
            "task_identifier": "SaveNexusPatternsAsAscii",
            "name": "enabled",
            "value": save_as_ascii,
        },
    ]

    inputs += task_inputs(
        task_identifier="SaveNexusPatternsAsId31TomoHdf5",
        inputs={
            "enabled": False,
            "scan_entry_url": "",
            "rot_name": "",
            "y_name": "",
            "output_filename": "",
        },
    )

    return {"inputs": inputs, "convert_destination": convert_destination}


def iter_scan_info(dataset_filename, detector_name, monitor_name=None):
    with h5py.File(dataset_filename, "r") as nxroot:
        for scan_name in nxroot:
            try:
                scan = nxroot[scan_name]
                if detector_name not in scan["measurement"]:
                    continue
                if monitor_name and monitor_name not in scan["measurement"]:
                    continue
                energy = scan[f"instrument/positioners/{ENERGY_POSITIONER}"][()]
            except Exception:
                print(f"Skip invalid scan {dataset_filename}::/{scan_name}")
                continue
            yield int(float(scan_name)), energy


def create_argument_parser():
    parser = argparse.ArgumentParser(
        description="",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "--beamline", type=str.lower, default="id31", help="Beamline name"
    )

    parser.add_argument(
        "--proposal", type=str.lower, required=True, help="Proposal name"
    )

    parser.add_argument(
        "--monitor",
        type=str,
        default="scaled_mondio",
        required=False,
        help="Flux monitor name",
    )

    parser.add_argument(
        "--reference", type=float, default=1, required=False, help="Reference counts"
    )

    parser.add_argument(
        "--samples",
        type=str,
        required=False,
        nargs="+",
        help="Samples to reprocess",
    )

    parser.add_argument(
        "-c",
        "--pyfai-config",
        type=str,
        required=True,
        help="PyFAI config file (.json)",
        metavar="FILE",
    )

    parser.add_argument(
        "--pyfai-method",
        type=str,
        default="full_csr_ocl_gpu",
        required=False,
        help="PyFAI integrator method",
    )

    parser.add_argument(
        "--worker",
        action="store_true",
        help="Execute workflows on ewoks worker instead of current environment",
    )

    parser.add_argument(
        "--save-ascii",
        action="store_true",
        help="Export integrated patterns as text files",
    )

    parser.add_argument(
        "--run",
        type=str,
        default="test",
        required=False,
        help="Processed sub-directory name",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Do not run anything but print which datasets would be processed",
    )

    return parser


def main(argv=None) -> None:
    if argv is None:
        argv = sys.argv

    parser = create_argument_parser()
    args = parser.parse_args(argv[1:])

    session_pattern = os.path.join(
        os.sep, "data", "visitor", args.proposal, args.beamline, "*"
    )
    sessions = glob(os.path.join(session_pattern))

    if len(sessions) > 1:
        session = input(f"Select session {sessions}: ")
    elif sessions:
        session = sessions[0]
    else:
        print("Proposal has no sessions")
        return

    samples = args.samples
    if not samples:
        sample_pattern = os.path.join(session, "RAW_DATA", "*", "*.h5")
        samples = [
            os.path.basename(os.path.dirname(sample_filename))
            for sample_filename in glob(sample_pattern)
        ]

    if args.worker:
        os.environ["BEACON_HOST"] = BEACON_HOST

    failed = list()
    futures = list()
    for sample in samples:
        dataset_pattern = os.path.join(session, "RAW_DATA", sample, "*", "*.h5")
        for dataset_filename in glob(dataset_pattern):
            scans = list(
                iter_scan_info(dataset_filename, DETECTOR, monitor_name=args.monitor)
            )
            for scan_number, energy in scans:
                kwargs = get_parameters(
                    dataset_filename,
                    scan_number,
                    DETECTOR,
                    args.pyfai_config,
                    args.pyfai_method,
                    energy,
                    monitor_name=args.monitor,
                    reference_counts=args.reference,
                    run=args.run,
                    save_as_ascii=args.save_ascii,
                )
                kwargs["load_options"] = WORKFLOW_LOAD_OPTIONS
                if args.dry_run:
                    print(f"Would process {dataset_filename}::/{scan_number}.1 with:")
                    print(json.dumps(kwargs, indent=2, sort_keys=True))
                elif args.worker:
                    future = submit(args=(WORKFLOW,), kwargs=kwargs)
                    futures.append((future, dataset_filename, scan_number))
                    print(f"Submitted {dataset_filename}::/{scan_number}.1")
                else:
                    try:
                        print(f"Processing {dataset_filename}::/{scan_number}.1 ...")
                        execute_graph(WORKFLOW, **kwargs)
                    except Exception as e:
                        traceback.print_exc()
                        failed.append(f"{dataset_filename}::/{scan_number}.1 ({e})")

    for future, dataset_filename, scan_number in futures:
        try:
            future.get()
        except Exception as e:
            print(f"FAILED {dataset_filename}::/{scan_number}.1")
            traceback.print_exc()
            failed.append(f"{dataset_filename}::/{scan_number}.1 ({e})")
        else:
            print(f"COMPLETED {dataset_filename}::/{scan_number}.1")

    if failed:
        print("FAILED:")
        print("\n".join(failed))


if __name__ == "__main__":
    sys.exit(main())



import numpy
from scipy.io import loadmat
from ewokscore import Task


class FlatFieldFromEnergy(
    Task,
    input_names=["newflat", "oldflat", "energy"],
    output_names=["flatfield"],
):
    """Interpolate an energy-stack of flat-field images.

    The resulting flat-field image can be used as follows
    to correct diffraction patterns for flat field:

    .. code::

        Icor = I / flatfield
    """

    def run(self):
        energy = self.inputs.energy
        new_flat = _interpolated_flatfield(self.inputs.newflat, "E", "F", energy)
        old_flat = _interpolated_flatfield(self.inputs.oldflat, "Eold", "Fold", energy)
        flatfield = old_flat / new_flat
        flatfield[~numpy.isfinite(flatfield)] = 1
        flatfield[flatfield < 0] = 1
        self.outputs.flatfield = flatfield


def _interpolated_flatfield(
    matlab_filename: str, energy_key: str, flatfield_key: str, energy: float
) -> numpy.ndarray:
    """
    :param matlab_filename: matlab file from ID31 that contains an energy-stack of images
    :param energy_key:
    :param flatfield_key:
    :param energy:
    :return: interpolated image (nrow, ncol)
    """
    m = loadmat(matlab_filename)
    flatfields = m[flatfield_key]
    energies = numpy.squeeze(m[energy_key]).astype(float)
    return _interpolate_flatfields(energies, flatfields, energy)


def _interpolate_flatfields(
    energies: numpy.ndarray, flatfields: numpy.ndarray, energy: float
) -> numpy.ndarray:
    """
    :param energies: array of energies (nenergies,)
    :param flatfields: stack of images (nrow, ncol, nenergies)
    :param energy:
    :return: interpolated image (nrow, ncol)
    """
    j = numpy.argsort(abs(energies - energy))[0:2]
    j.sort()
    j_before, j_after = j
    e_before = energies[j_before]
    e_after = energies[j_after]
    w_before = (e_after - energy) / (e_after - e_before)
    w_after = (energy - e_before) / (e_after - e_before)
    f_before = flatfields[..., j_before]
    f_after = flatfields[..., j_after]
    return w_before * f_before + w_after * f_after



from __future__ import annotations

import hdf5plugin  # noqa
import h5py

import silx.gui.hdf5
from silx.gui.hdf5._utils import H5Node
from silx.gui import qt
from silx.gui.data.DataViewerFrame import DataViewerFrame


class Hdf5TreeView(qt.QMainWindow):

    clicked = qt.Signal()

    def __init__(self):
        super().__init__()
        self.__loadedFiles: set[str] = set()

        self.__treeview = silx.gui.hdf5.Hdf5TreeView(self)
        self.__treeview.setSelectionMode(
            qt.QAbstractItemView.SelectionMode.ExtendedSelection
        )

        self.__dataViewer = DataViewerFrame(self)

        self.__treeview.setMinimumWidth(100)
        self.__dataViewer.setMinimumWidth(100)

        splitter = qt.QSplitter(self)
        splitter.addWidget(self.__treeview)
        splitter.addWidget(self.__dataViewer)

        splitter.setCollapsible(0, False)
        splitter.setCollapsible(1, False)

        splitter.setHandleWidth(10)
        splitter.setStretchFactor(1, 1)

        centralWidget = qt.QWidget(self)
        layout = qt.QVBoxLayout(centralWidget)
        layout.addWidget(splitter)

        self.setCentralWidget(centralWidget)

        self.__treeview.activated.connect(self.__displayData)
        self.__treeview.addContextMenuCallback(self.__closeAndSyncCustomContextMenu)
        self.__treeview.clicked.connect(self.clicked)

    def __displayData(self) -> None:
        """
        Update the data viewer with the currently selected HDF5 node.
        """
        selected = list(self.__treeview.selectedH5Nodes())
        if len(selected) >= 1:
            data = selected[-1]
            self.__dataViewer.setData(data)

    def addFile(self, filename: str) -> None:
        """
        Add a file to the HDF5 tree viewer.

        Prevents adding duplicate files by checking if the file is already loaded.

        Args:
            filename: Path to the HDF5 file to add.
        """
        if filename in self.__loadedFiles:
            qt.QMessageBox.information(
                self,
                "Duplicate File",
                f"This file {filename} is already loaded.",
            )
            return

        self.__treeview.findHdf5TreeModel().appendFile(filename)
        self.__loadedFiles.add(filename)

    def clearFiles(self) -> None:
        """
        Remove and close all files currently loaded in the HDF5 tree viewer.

        Resets the internal file tracking list.
        """
        self.__treeview.findHdf5TreeModel().clear()
        self.__loadedFiles.clear()

    def getLoadedFiles(self) -> set[str]:
        """
        Get the set of currently loaded files in the HDF5 tree viewer.

        Returns:
            A set containing the file paths of all loaded files.
        """
        return self.__loadedFiles

    def isEmpty(self) -> bool:
        """
        Check if the HDF5 tree viewer contains any loaded files.

        Returns:
            True if no files are currently loaded, False otherwise.
        """
        return len(self.__loadedFiles) == 0

    def reloadSelected(self) -> None:
        """
        Reload the currently selected HDF5 nodes.
        """
        model = self.__treeview.findHdf5TreeModel()
        for obj in self.getSelectedNodes():
            if obj.ntype is h5py.File:
                model.synchronizeH5pyObject(obj.h5py_object)

    def getSelectedNodes(self) -> list[H5Node]:
        """
        Return a list of the currently selected HDF5 nodes in the tree view.
        """
        return list(self.__treeview.selectedH5Nodes())

    def __closeAndSyncCustomContextMenu(
        self, event: silx.gui.hdf5.Hdf5ContextMenuEvent
    ) -> None:
        """
        Populate the custom context menu for the HDF5 tree viewer.

        Adds additional options to remove or reload HDF5 files in the tree view when the
        context menu is triggered.
        """
        selectedObjects = event.source().selectedH5Nodes()
        menu = event.menu()

        if not menu.isEmpty():
            menu.addSeparator()

        for obj in selectedObjects:
            if obj.ntype is h5py.File:
                action = qt.QAction("Remove %s" % obj.local_filename, event.source())
                action.triggered.connect(
                    lambda: self.__treeview.findHdf5TreeModel().removeH5pyObject(
                        obj.h5py_object
                    )
                )
                menu.addAction(action)
                action = qt.QAction("Reload %s" % obj.local_filename, event.source())
                action.triggered.connect(
                    lambda: self.__treeview.findHdf5TreeModel().synchronizeH5pyObject(
                        obj.h5py_object
                    )
                )
                menu.addAction(action)



from __future__ import annotations

import argparse
import os
import logging
import threading
import sys
import ctypes
from importlib.metadata import version

# Environment variable HDF5_USE_FILE_LOCKING must be set *before* importing HDF5 libraries
# This prevents file locking issues, especially in shared filesystems (e.g., NFS).
os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"

from .hdf5widget import Hdf5TreeView  # noqa: E402
from .utils import (  # noqa: E402
    generateInputs,
    extractScanNumber,
    getScanEnergy,
    FilenameCompleterLineEdit,
    get_scan_url,
)
from ewoks.bindings import execute_graph  # noqa: E402
from ewoks.bindings import submit_graph  # noqa: E402

import PyQt5.QtCore  # noqa: F401,E402; Needed to force PyQt5
from silx.gui import qt, icons, hdf5  # noqa: E402
from silx.gui.utils.concurrent import submitToQtMainThread  # noqa: E402

from ..utils import (  # noqa: E402
    FLATFIELD_DEFAULT_DIR,
    NEWFLAT_FILENAME,
    OLDFLAT_FILENAME,
)
from .output_widget import OutputWidget  # noqa: E402

_logger = logging.getLogger(__name__)

_FILE_EXTENSIONS = ".h5", ".hdf5", ".hdf", ".nx", ".nxs", ".nx5", ".nexus"


class Id31FAIEwoksMainWindow(qt.QMainWindow):
    """
    This GUI is designed for reprocessing HDF5 scans data with fast azimuthal
    integration using Ewoks workflows (ewoksXRPD)
    """

    _SETTINGS_VERSION_STR: str = "1"
    _DETECTOR_NAMES: tuple[str] = (
        "p3",
        "de",
        "perkin",
    )
    _MONITOR_NAMES: tuple[str] = "mondio", "scaled_detdio", "scaled_mondio", "srcur"
    _INTEGRATION_METHODS: tuple[str] = (
        "no_csr_cython",
        "bbox_csr_cython",
        "full_csr_cython",
        "no_csr_ocl_gpu",
        "bbox_csr_ocl_gpu",
        "full_csr_ocl_gpu",
        "no_histogram_cython",
        "bbox_histogram_cython",
        "full_histogram_ocl_gpu",
    )

    _WORKFLOW_WITH_FLAT: str = "integrate_with_saving_with_flat.json"
    _WORKFLOW_WITHOUT_FLAT: str = "integrate_with_saving.json"
    _WORKFLOW_LOAD_OPTIONS: dict = {"root_module": "ewoksid31.workflows"}
    _HOSTS: tuple[str] = "Local", "Ewoks worker"

    def __init__(self, parent: qt.QWidget | None = None):
        super().__init__(parent)
        self.setWindowTitle("pyFAI with EWOKS - ID31")
        self.resize(1000, 800)

        self.statusBar = qt.QStatusBar()
        self.setStatusBar(self.statusBar)

        # Set paths
        self._defaultDirectoryRaw = ""
        self._flatFieldDirName = FLATFIELD_DEFAULT_DIR

        # Central Layout setup
        centralWidget = qt.QWidget(self)
        self.setCentralWidget(centralWidget)
        mainLayout = qt.QVBoxLayout(centralWidget)

        # HDF5 Viewer
        self._hdf5Widget = Hdf5TreeView()
        mainLayout.addWidget(self._hdf5Widget)

        # Menu and Toolbar setup
        self._setupMenuAndToolBar()

        # pyFAI Config Section
        pyFaiGroupBox = qt.QGroupBox("pyFAI config")
        pyFaiLayout = qt.QGridLayout()

        self._configFileLineEdit = FilenameCompleterLineEdit()
        self._configFileLineEdit.setPlaceholderText("/path/to/pyfai_config.json")

        loadConfigFileButton = qt.QPushButton("Open...")
        loadConfigFileButton.setIcon(icons.getQIcon("document-open"))
        loadConfigFileButton.clicked.connect(self._loadConfigFileButtonClicked)

        pyFaiLayout.addWidget(self._configFileLineEdit, 0, 0, 1, 1)
        pyFaiLayout.addWidget(loadConfigFileButton, 0, 1, 1, 1)

        pyFaiGroupBox.setLayout(pyFaiLayout)
        mainLayout.addWidget(pyFaiGroupBox)

        # Processing Options Section
        processingOptionsGroupBox = qt.QGroupBox("Processing options")
        processingOptionsLayout = qt.QGridLayout()

        spacer = qt.QSpacerItem(
            20, 10, qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        processingOptionsLayout.addItem(spacer, 0, 4, 2, 4)

        detectorLabel = qt.QLabel("Detector:")
        self._detectorComboBox = qt.QComboBox()
        self._detectorComboBox.addItems(self._DETECTOR_NAMES)
        self._detectorComboBox.setCurrentText("p3")
        self._detectorComboBox.setToolTip("Select the detector type.")
        self._detectorComboBox.activated.connect(self._onDetectorChanged)

        processingOptionsLayout.addWidget(detectorLabel, 0, 0, 1, 1, qt.Qt.AlignLeft)
        processingOptionsLayout.addWidget(self._detectorComboBox, 0, 1, 1, 2)

        monitorNameLabel = qt.QLabel("Monitor name:")
        self._monitorNameComboBox = qt.QComboBox()
        self._monitorNameComboBox.addItems(self._MONITOR_NAMES)
        self._monitorNameComboBox.setCurrentText("scaled_mondio")
        self._monitorNameComboBox.setToolTip(
            "Select the monitor name for normalization."
        )

        processingOptionsLayout.addWidget(monitorNameLabel, 1, 0, 1, 1, qt.Qt.AlignLeft)
        processingOptionsLayout.addWidget(self._monitorNameComboBox, 1, 1, 1, 2)

        self._sigmaClipCheckBox = qt.QCheckBox("Sigma clipping threshold:")
        self._sigmaClipCheckBox.setChecked(False)
        self._sigmaClipCheckBox.setToolTip(
            "Check to enable sigma clipping and set threshold."
        )

        self._sigmaClipThresholdSpinBox = qt.QDoubleSpinBox()
        self._sigmaClipThresholdSpinBox.setSingleStep(0.1)
        self._sigmaClipThresholdSpinBox.setRange(0.1, 10.0)
        self._sigmaClipThresholdSpinBox.setValue(3.0)
        self._sigmaClipThresholdSpinBox.setDecimals(1)
        self._sigmaClipThresholdSpinBox.setEnabled(False)

        self._sigmaClipCheckBox.toggled.connect(
            self._sigmaClipThresholdSpinBox.setEnabled
        )
        self._sigmaClipCheckBox.toggled.connect(self._sigmaClipCheckBoxToggled)

        processingOptionsLayout.addWidget(
            self._sigmaClipCheckBox, 2, 0, 1, 1, qt.Qt.AlignLeft
        )
        processingOptionsLayout.addWidget(
            self._sigmaClipThresholdSpinBox, 2, 1, 1, 2, qt.Qt.AlignLeft
        )

        integrationMethodLabel = qt.QLabel("Integration method:")
        self._integrationMethodComboBox = qt.QComboBox()
        self._integrationMethodComboBox.addItems(self._INTEGRATION_METHODS)
        self._integrationMethodComboBox.setCurrentText("no_histogram_cython")
        self._integrationMethodComboBox.setToolTip(
            "Select the pyFAI integration method."
        )

        processingOptionsLayout.addWidget(
            integrationMethodLabel, 3, 0, 1, 1, qt.Qt.AlignLeft
        )
        processingOptionsLayout.addWidget(self._integrationMethodComboBox, 3, 1, 1, 2)

        radialUnitLabel = qt.QLabel("Radial unit:")
        self._qCheckBox = qt.QCheckBox("q (scattering vector)")
        self._qUnitsComboBox = qt.QComboBox()
        self._qUnitsComboBox.addItems(("Å⁻¹", "nm⁻¹"))
        self._2thCheckBox = qt.QCheckBox("2theta (degree)")
        self._qUnitsComboBox.setEnabled(False)
        self._qCheckBox.toggled.connect(self._qUnitsComboBox.setEnabled)

        processingOptionsLayout.addWidget(radialUnitLabel, 4, 0, 1, 1, qt.Qt.AlignLeft)
        processingOptionsLayout.addWidget(self._qCheckBox, 4, 1, 1, 1, qt.Qt.AlignLeft)
        processingOptionsLayout.addWidget(
            self._qUnitsComboBox, 4, 2, 1, 1, qt.Qt.AlignRight
        )
        processingOptionsLayout.addWidget(
            self._2thCheckBox, 5, 1, 1, 1, qt.Qt.AlignLeft
        )

        processingOptionsGroupBox.setLayout(processingOptionsLayout)
        mainLayout.addWidget(processingOptionsGroupBox)

        # Output Section
        self._outputWidget = OutputWidget()
        mainLayout.addWidget(self._outputWidget)

        # Run Section
        executionModeLayout = qt.QHBoxLayout()

        executionModeLabel = qt.QLabel("Host:")
        self._executionModeComboBox = qt.QComboBox()
        self._executionModeComboBox.addItems(self._HOSTS)
        self._executionModeComboBox.setCurrentText("Local")
        self._executionModeComboBox.setToolTip(
            "Select where the process will be executed: locally or on an Ewoks worker."
        )

        runButton = qt.QPushButton("Run")
        runButton.setIcon(icons.getQIcon("next"))
        runButton.clicked.connect(self._runButtonClicked)
        runButton.setToolTip("Run the processing workflow on the selected host.")
        runButton.setMinimumSize(runButton.sizeHint() * 2)

        executionModeLayout.addStretch()
        executionModeLayout.addWidget(runButton)
        executionModeLayout.addWidget(executionModeLabel)
        executionModeLayout.addWidget(self._executionModeComboBox)
        executionModeLayout.addStretch()

        mainLayout.addLayout(executionModeLayout)

    def _showHelpDialogClicked(self):
        """
        Display the Help dialog when the Help button is clicked.
        """
        helpDialog = qt.QMessageBox(self)
        helpDialog.setWindowTitle("Help")
        helpDialog.setIcon(qt.QMessageBox.Information)

        helpDialog.setTextFormat(qt.Qt.RichText)
        helpDialog.setText(
            f"<b>Welcome to the pyFAI with EWOKS application help (version {version('ewoksid31')}).</b><br>"
            "<br>How to get started?<br>"
            "<ul>"
            f"<li>Load raw data files {_FILE_EXTENSIONS}.</li>"
            "<li>Load the pyFAI configuration file.</li>"
            "<li>Select an output directory to save the processed data.</li>"
            "<li>Click Run to execute the workflow.</li>"
            "</ul>"
            "<br>For more information, visit:<br>"
            '<a href="https://ewoksid31.readthedocs.io/en/latest/">Ewoksid31 Documentation</a><br>'
            '<a href="https://confluence.esrf.fr/display/ID31KB/GUI+for+reprocessing+XRPD+data">ID31 GUI Confluence Page</a>'
        )

        helpDialog.setStandardButtons(qt.QMessageBox.Ok)
        helpDialog.exec_()

    def _setupMenuAndToolBar(
        self,
    ) -> None:
        """
        Setup of the main menu and toolbar with actions for file handling.
        """
        menuBar = self.menuBar()
        fileMenu = menuBar.addMenu("File")

        openAction = qt.QAction(icons.getQIcon("document-open"), "Open...", self)
        openAction.triggered.connect(self._openActionTriggered)
        openAction.setShortcut(qt.QKeySequence.StandardKey.Open)
        reloadAction = qt.QAction(icons.getQIcon("view-refresh"), "Reload", self)
        reloadAction.triggered.connect(self._hdf5Widget.reloadSelected)
        reloadAction.setShortcut(qt.QKeySequence.StandardKey.Refresh)
        clearAction = qt.QAction(icons.getQIcon("remove"), "Clear", self)
        clearAction.triggered.connect(self._hdf5Widget.clearFiles)
        clearAction.setShortcut(qt.QKeySequence.StandardKey.Delete)

        fileMenu.addActions([openAction, reloadAction, clearAction])

        toolBar = self.addToolBar("File Actions")
        toolBar.addActions([openAction, reloadAction, clearAction])
        toolBar.setMovable(False)
        self.setContextMenuPolicy(qt.Qt.PreventContextMenu)
        toolBar.setFloatable(False)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Preferred)
        toolBar.addWidget(spacer)

        helpAction = qt.QAction(
            self.style().standardIcon(qt.QStyle.SP_TitleBarContextHelpButton), "", self
        )
        font = helpAction.font()
        font.setBold(True)
        helpAction.setFont(font)
        helpAction.setToolTip("About this app")
        helpAction.triggered.connect(self._showHelpDialogClicked)
        toolBar.addAction(helpAction)

    def addRawDataFile(self, fileName: str) -> None:
        """
        Proxy method to add a new file to the HDF5 tree viewer.
        """
        if not fileName or not os.path.isfile(fileName):
            qt.QMessageBox.warning(
                self, "Invalid File Format", "The selected file does not exist."
            )
            return
        if not fileName.endswith(_FILE_EXTENSIONS):
            qt.QMessageBox.warning(
                self,
                "Invalid File Format",
                f"Please select a valid HDF5 or NEXUS file {_FILE_EXTENSIONS}.",
            )
            return
        self._hdf5Widget.addFile(fileName)
        self._defaultDirectoryRaw = os.path.dirname(fileName)

    def _openActionTriggered(self) -> None:
        """
        Add Raw data as HDF5 file without cleaning the tree viewer.
        """
        nameFilter = " ".join(f"*{ext}" for ext in _FILE_EXTENSIONS)
        fileName, _ = qt.QFileDialog.getOpenFileName(
            self,
            "Add RAW data file",
            self._defaultDirectoryRaw,
            f"HDF5 files ({nameFilter});;All files (*)",
        )
        if fileName:
            self.addRawDataFile(fileName)

    def _loadConfigFileButtonClicked(self) -> None:
        """
        Choose and import JSON or PONI config file.
        """
        currentDir = os.path.dirname(self.getConfigFilePath()) or os.getcwd()
        filePath, _ = qt.QFileDialog.getOpenFileName(
            self,
            "Open config file",
            currentDir,
            "JSON or PONI files (*.json *.poni);;All files (*)",
        )
        if filePath:
            self.setConfigFilePath(filePath)
        else:
            self.setConfigFilePath("")
            _logger.info("No config file chosen.")

    def getFlatFieldDirName(self) -> str:
        """Returns the flat-field directory where to find flats.mat and oldflats.mat"""
        return self._flatFieldDirName

    def setFlatFieldDirName(self, path: str):
        """Set the directory where to find flats.mat and oldflats.mat"""
        self._flatFieldDirName = os.path.abspath(path)

    def getConfigFilePath(self) -> str:
        """
        Returns the current configuration file path from the line edit.
        """
        return self._configFileLineEdit.text().strip()

    def setConfigFilePath(self, path: str) -> None:
        """
        Update the configuration file path in the line edit.
        """
        self._configFileLineEdit.setText(path)

    def _sigmaClipCheckBoxToggled(self, toggled: bool) -> None:
        """
        Update integration methods combo box according to sigma clipping.

        - Change selected method if not supported by sigma clipping
        - Enable/Disable pixel splitting methods depending on sigma clipping
        """
        noCsrPrefix = "no_csr"

        if toggled and not self._integrationMethodComboBox.currentText().startswith(
            noCsrPrefix
        ):
            self._integrationMethodComboBox.setCurrentText("no_csr_ocl_gpu")

        model = self._integrationMethodComboBox.model()
        for row in range(self._integrationMethodComboBox.count()):
            item = model.item(row)
            if not item.text().startswith(noCsrPrefix):
                flags = item.flags()
                if toggled:
                    item.setFlags(flags & ~qt.Qt.ItemIsEnabled)
                    item.setToolTip("Not available with sigma clipping")
                else:
                    item.setFlags(flags | qt.Qt.ItemIsEnabled)
                    item.setToolTip("")

    def _getSigmaClippingThreshold(self) -> float | None:
        """
        Retrieve Sigma Clipping threshold from the user interface.

        Returns:
            The Sigma Clipping threshold if enabled.
        """
        if self._sigmaClipCheckBox.isChecked():
            return float(self._sigmaClipThresholdSpinBox.value())
        return None

    def _onDetectorChanged(self) -> None:
        """
        Prompt the user to verify the configuration file after changing the detector.

        Ensures that the selected config file in the 'pyFAI config' section is compatible with the newly selected detector.
        """

        selectedDetector = self._detectorComboBox.currentText()

        message = (
            f"You just changed the detector to <b><i>{selectedDetector}</i></b>.<br><br>"
            "Please ensure that the <b>config file</b> in the <i>pyFAI config</i> section is correct.<br>"
            "If necessary, load the appropriate <b>JSON</b> or <b>PONI</b> file."
        )

        qt.QMessageBox.information(self, "Configuration file verification", message)

    def _getSelectedUnits(self) -> list[str]:
        """
        Retrieves the selected units for azimuthal integration.

        Returns:
            A list of selected units, including q (Å⁻¹ or nm⁻¹) and/or 2θ (degrees).
        """
        units = list()
        if self._qCheckBox.isChecked():
            if self._qUnitsComboBox.currentText() == "Å⁻¹":
                units.append("q_A^-1")
            elif self._qUnitsComboBox.currentText() == "nm⁻¹":
                units.append("q_nm^-1")
        if self._2thCheckBox.isChecked():
            units.append("2th_deg")
        return units

    def _getParameters(
        self, datasetFilename: str, scanNumber: int, unit: str
    ) -> dict | None:
        """
        Generates parameters to execute workflow for a given scan.

        Args:
            datasetFilename: Filename of the HDF5 file containing the scan
            scanNumber: Number of the scan.
            unit: Selected unit.

        Returns:
            Dictionnary of parameters for the workflow, or None if failed.
        """
        energy = getScanEnergy(datasetFilename, scanNumber)

        if energy is None:
            _logger.error(
                f"Unable to retrieve energy for scan {scanNumber} in file {datasetFilename}."
            )
            return None

        inputParameters = generateInputs(
            newFlat=os.path.join(self.getFlatFieldDirName(), NEWFLAT_FILENAME),
            oldFlat=os.path.join(self.getFlatFieldDirName(), OLDFLAT_FILENAME),
            energy=energy,
            pyfaiConfig=self.getConfigFilePath(),
            pyfaiMethod=self._integrationMethodComboBox.currentText(),
            datasetFilename=datasetFilename,
            scanNumber=scanNumber,
            monitorName=self._monitorNameComboBox.currentText(),
            referenceCounts=1,
            detectorName=self._detectorComboBox.currentText(),
            outputDirectory=self._outputWidget.getOutputDirName(),
            sigmaClippingThreshold=self._getSigmaClippingThreshold(),
            asciiExport=self._outputWidget.isAsciiExportEnabled(),
            unit=unit,
            tomoEnabled=self._outputWidget.isTomoEnabled(),
            tomoScanEntryUrl=f"{datasetFilename}::/{scanNumber}.1",
            rotName=self._outputWidget.getTomoRotName(),
            yName=self._outputWidget.getTomoYName(),
            tomoOutputFilename=os.path.join(
                self._outputWidget.getOutputDirName(), f"{scanNumber}_tomo.h5"
            ),
        )

        return inputParameters

    def _prepareScans(self, scans: list[hdf5.H5Node]) -> list[dict]:
        """
        Prepares parameters for each unique selected scan and selected unit.
        """
        uniqueScans = set()
        for scan in scans:
            scanNumber = extractScanNumber(scan.physical_name)
            if scanNumber == -1:
                _logger.warning(
                    f"Skipping scan: Invalid scan number in {scan.physical_name}"
                )
                continue
            uniqueScans.add((scan.physical_filename, scanNumber))

        inputParameters = list()
        selectedUnits = self._getSelectedUnits()
        for rawDataFilename, scanNumber in uniqueScans:
            for unit in selectedUnits:
                workflowParameters = self._getParameters(
                    rawDataFilename, scanNumber, unit
                )
                if not workflowParameters:
                    _logger.warning(
                        f"Skipping scan {rawDataFilename}::{scanNumber} due to missing parameters."
                    )
                    continue

                inputParameters.append(workflowParameters)

        return inputParameters

    def _processScans(self, scans: list[hdf5.H5Node], local: bool = True) -> None:
        """
        Execute workflow and save HDF5 data and JSON workflow ewoks file from a single selected scan
        and selected unit.

        Executing in a separate thread.

        Args:
            scans: Selected scan nodes
            local: Whether to execute locally or submit to Ewoks worker.
        """
        inputParameters = self._prepareScans(scans)

        if local:
            self._updateStatusBar("Processing scans locally...")
            _logger.info("Processing scans locally...")
            thread = threading.Thread(
                target=self._executeWorkflowForParams, args=(inputParameters,)
            )
            thread.start()
        else:
            self._updateStatusBar("Submitting scans to Ewoks worker...")
            _logger.info("Submitting scans to Ewoks worker...")
            self._submitWorkflowForParams(inputParameters)

    def _showErrorInStatusBar(self, message: str) -> None:
        """
        Display an error message in the status bar and log it.

        Args:
            message: The error message to display.
        """
        _logger.error(message)
        submitToQtMainThread(self._updateStatusBar, message, error=True)

    def _getWorkflow(self) -> str:
        """
        Returns the appropriate workflow based on the selected detector.
        """
        if self._detectorComboBox.currentText() == "p3":
            return self._WORKFLOW_WITH_FLAT
        return self._WORKFLOW_WITHOUT_FLAT

    def _executeWorkflowForParams(self, params: list[dict]) -> None:
        """
        Workflow execution for multiple scans in a single thread (locally).

        Args:
            params: List of workflow input parameters for each scan.
        """
        for workflowParameters in params:
            datasetFilename, scanNumber = get_scan_url(workflowParameters)

            try:
                execute_graph(
                    self._getWorkflow(),
                    load_options=self._WORKFLOW_LOAD_OPTIONS,
                    **workflowParameters,
                )
                _logger.info(
                    f"Successfully processed scan: {os.path.basename(datasetFilename)}::{scanNumber}"
                )
            except Exception as e:
                errorMsg = f"Error processing scan {scanNumber}: {e}"
                _logger.error(errorMsg)
                submitToQtMainThread(self._updateStatusBar, errorMsg, error=True)
                raise e

        submitToQtMainThread(
            self._updateStatusBar, "Processing completed successfully."
        )
        _logger.info(f"All results saved in: {self._outputWidget.getOutputDirName()}")

    def _runButtonClicked(self) -> None:
        """
        Handle the run button click. Validate inputs, adjust output directory and process scans based on the selected execution mode.
        """
        if not self._validateInputParameters():
            return
        pass

        selectedScanNodes = list(self._hdf5Widget.getSelectedNodes())

        if not self._adjustOutputDirectory():
            self._updateStatusBar("Process canceled by user.", error=True)
            return

        executionMode = self._executionModeComboBox.currentText()

        if executionMode == "Local":
            self._processScans(selectedScanNodes, local=True)
        elif executionMode == "Ewoks worker":
            self._processScans(selectedScanNodes, local=False)
        else:
            self._updateStatusBar(
                f"Unknown execution mode: {executionMode}.", error=True
            )

    def _processFutures(self, futures: list[tuple]) -> None:
        """
        Process the futures in a separate thread.

        Args:
            futures: A list of tuples containing -> (future, scanName).
        """
        failed = list()
        for future, scanName in futures:
            try:
                future.get()
            except Exception as e:
                _logger.error(f"Workflow failed for {scanName}: {e}")
                failed.append(scanName)

        if failed:
            errorMsg = f"Workflows completed with errors for scans: {', '.join(failed)}"
            submitToQtMainThread(self._updateStatusBar, errorMsg, error=True)

        else:
            successMsg = "All workflows completed successfully."
            submitToQtMainThread(self._updateStatusBar, successMsg)
            _logger.info(
                f"All results saved in: {self._outputWidget.getOutputDirName()}"
            )

    def _submitWorkflowForParams(self, params: list[dict]) -> None:
        """
        Submit workflow for multiple scans to the Ewoks worker (remotely).

        Args:
            params: List of workflow input parameters for each scan.
        """
        futures = list()

        for workflowParameters in params:
            scanName = workflowParameters.pop("scanName", "Unknown scan")
            _logger.info(f"Submitting workflow for parameters: {scanName}")

            try:
                future = submit_graph(
                    self._getWorkflow(),
                    load_options=self._WORKFLOW_LOAD_OPTIONS,
                    **workflowParameters,
                )
                futures.append((future, scanName))
                _logger.info(f"Successfully submitted scan: {scanName}")
            except Exception as e:
                errorMsg = f"Error submitting scan {scanName}: {e}"
                _logger.error(errorMsg)
                self._updateStatusBar(errorMsg, error=True)

        thread = threading.Thread(target=self._processFutures, args=(futures,))
        thread.start()

    def _submitButtonClicked(self) -> None:
        """
        Handle the submit button click. Validate inputs, adjust output directory and submit scans to the worker.
        """

        if not self._validateInputParameters():
            return

        selectedScanNodes = list(self._hdf5Widget.getSelectedNodes())

        if not self._adjustOutputDirectory():
            self._updateStatusBar("Process canceled by user.", error=True)
            return

        self._processScans(selectedScanNodes, local=False)

    def _getValidationErrors(
        self,
        rawDataLoaded: bool,
        selectedScanNodes: list[hdf5.H5Node],
        configFileLoaded: bool,
        unitSelected: bool,
    ) -> list[str]:
        """
        Generate a list of validation error messages based on the current state.

        Args:
            rawDataLoaded: Whether raw data is loaded.
            selectedScanNodes: List of selected scan nodes.
            configFileLoaded: Whether a configuration file is loaded.
            unitSelected: Whether at least one radial unit is selected.

        Returns:
            A list of error messages if validation fails, otherwise an empty list.
        """
        errors = list()

        isOnlyRootSelected = (
            len(selectedScanNodes) == 1 and selectedScanNodes[0].physical_name == "/"
        )

        if isOnlyRootSelected:
            errors.append(
                "Only the root of the HDF5 file is selected. Please select one or more scans."
            )

        if not rawDataLoaded:
            errors.append("No raw data file loaded.")
        if not selectedScanNodes:
            errors.append("No scan selected.")
        if not configFileLoaded:
            errors.append("No pyFAI config file loaded.")
        if not unitSelected:
            errors.append("No radial unit selected.")
        return errors

    def _showWarningMessage(self, errorMessages: list[str]) -> None:
        """
        Show a warning message box if prerequisites for processing are not met.
        """
        warningMessageBox = qt.QMessageBox(self)
        warningMessageBox.setWindowTitle("Workflow cannot be excecuted")
        warningMessageBox.setIcon(qt.QMessageBox.Warning)
        warningMessageBox.setText("\n".join(errorMessages))
        warningMessageBox.exec_()

    def _validateInputParameters(self) -> bool:
        """
        Validates that all necessary inputs are present before starting processing.

        Returns:
            True if inputs are valid, False otherwise.
        """
        rawDataLoaded = not self._hdf5Widget.isEmpty()
        selectedScanNodes = self._hdf5Widget.getSelectedNodes()
        configFileLoaded = bool(self.getConfigFilePath())

        unitSelected = self._qCheckBox.isChecked() or self._2thCheckBox.isChecked()

        errorMessages = self._getValidationErrors(
            rawDataLoaded, selectedScanNodes, configFileLoaded, unitSelected
        )

        if errorMessages:
            self._updateStatusBar(" ".join(errorMessages), error=True)
            _logger.warning(f"Processing cannot start: {' '.join(errorMessages)}")
            self._showWarningMessage(errorMessages)
            return False
        return True

    def _showConfirmationMessage(self, outputDirectory: str) -> bool:
        """
        Show a confirmation message box when creating a new directory for output file.

        Returns True is the user confirms, otherwise False.
        """
        confirmationMessageBox = qt.QMessageBox(self)
        confirmationMessageBox.setWindowTitle("Overwrite Output Files")
        confirmationMessageBox.setIcon(qt.QMessageBox.Question)
        confirmationMessageBox.setText("The selected output directory already exists.")
        confirmationMessageBox.setInformativeText(
            f"Do you want to overwrite existing files in: {outputDirectory} ?\nUnchanged files will be preserved."
        )
        confirmationMessageBox.setStandardButtons(
            qt.QMessageBox.Ok | qt.QMessageBox.Cancel
        )
        userResponse = confirmationMessageBox.exec_()
        return userResponse == qt.QMessageBox.Ok

    def _adjustOutputDirectory(self) -> bool:
        """
        Use the selected output directory directly, without creating a new one.
        Ask for confirmation if the folder already exists.

        Returns:
            True if the process should continue, False if canceled.
        """
        currentOutputDir = self._outputWidget.getOutputDirName()
        if os.path.exists(currentOutputDir):
            if not self._showConfirmationMessage(currentOutputDir):
                _logger.info("Process canceled by the user.")
                return False
            _logger.info(
                f"Processed data will be written in existing directory: {currentOutputDir}"
            )

        return True

    def _updateStatusBar(self, message: str, error: bool = False):
        """
        Updates the status bar with a message.

        If 'error' is True, displays the message in red.
        """
        self.statusBar.setStyleSheet("color: red;" if error else "")
        self.statusBar.showMessage(message)

    def loadSettings(self) -> None:
        """
        Load user settings.
        """
        settings = qt.QSettings()

        if settings.value("version") != self._SETTINGS_VERSION_STR:
            _logger.info("Setting version mismatch. Clearing settings.")
            settings.clear()
            return

        geometry = settings.value("mainWindow/geometry")
        if geometry:
            self.restoreGeometry(geometry)

        self._defaultDirectoryRaw = settings.value("input/inputDirectory", "")

        self.setConfigFilePath(settings.value("config/configFile", ""))

        detectorName = settings.value("options/detectorName", None)
        if detectorName in self._DETECTOR_NAMES:
            self._detectorComboBox.setCurrentText(detectorName)

        monitorName = settings.value("options/monitorName", None)
        if monitorName in self._MONITOR_NAMES:
            self._monitorNameComboBox.setCurrentText(monitorName)

        self._sigmaClipCheckBox.setChecked(
            settings.value("options/sigmaClipping/enabled", False, type=bool)
        )
        self._sigmaClipThresholdSpinBox.setValue(
            settings.value("options/sigmaClipping/threshold", 3, type=float)
        )

        integrationMethod = settings.value("options/integrationMethod", None)
        if integrationMethod in self._INTEGRATION_METHODS:
            self._integrationMethodComboBox.setCurrentText(integrationMethod)

        self._qCheckBox.setChecked(settings.value("options/units/q", False, type=bool))
        selectedQUnit = settings.value("options/units/q_unit", "Å⁻¹")
        if selectedQUnit in ["Å⁻¹", "nm⁻¹"]:
            self._qUnitsComboBox.setCurrentText(selectedQUnit)

        self._2thCheckBox.setChecked(
            settings.value("options/units/2th", False, type=bool)
        )

        self._outputWidget.setTomoEnabled(
            settings.value("output/tomo/enabled", False, type=bool)
        )
        self._outputWidget.setTomoRotName(settings.value("output/tomo/rot_name", "nth"))
        self._outputWidget.setTomoYName(settings.value("output/tomo/y_name", "ny"))

        outputDirectory = settings.value("output/outputDirectory", "")
        self._outputWidget.setOutputDirName(outputDirectory)
        self._outputWidget.setAsciiExportEnabled(
            settings.value("output/asciiExport", False, type=bool)
        )

        executionMode = settings.value("run/executionMode", None)
        if executionMode in self._HOSTS:
            self._executionModeComboBox.setCurrentText(executionMode)

    def closeEvent(self, event: qt.QCloseEvent) -> None:
        """
        Save user settings on application close.
        """
        settings = qt.QSettings()

        settings.setValue("version", self._SETTINGS_VERSION_STR)
        settings.setValue("mainWindow/geometry", self.saveGeometry())

        settings.setValue("input/inputDirectory", self._defaultDirectoryRaw)

        settings.setValue("config/configFile", self.getConfigFilePath())

        settings.setValue("options/detectorName", self._detectorComboBox.currentText())
        settings.setValue(
            "options/monitorName", self._monitorNameComboBox.currentText()
        )
        settings.setValue(
            "options/sigmaClipping/enabled", self._sigmaClipCheckBox.isChecked()
        )
        settings.setValue(
            "options/sigmaClipping/threshold",
            float(self._sigmaClipThresholdSpinBox.value()),
        )
        settings.setValue(
            "options/integrationMethod", self._integrationMethodComboBox.currentText()
        )
        settings.setValue("options/units/q", self._qCheckBox.isChecked())
        settings.setValue("options/units/q_unit", self._qUnitsComboBox.currentText())
        settings.setValue("options/units/2th", self._2thCheckBox.isChecked())

        settings.setValue("output/tomo/enabled", self._outputWidget.isTomoEnabled())
        settings.setValue("output/tomo/rot_name", self._outputWidget.getTomoRotName())
        settings.setValue("output/tomo/y_name", self._outputWidget.getTomoYName())

        settings.setValue(
            "output/outputDirectory", self._outputWidget.getOutputDirName()
        )
        settings.value("output/asciiExport", self._outputWidget.isAsciiExportEnabled())

        settings.setValue(
            "run/executionMode", self._executionModeComboBox.currentText()
        )

        event.accept()


def create_argument_parser():
    parser = argparse.ArgumentParser(
        description="",
    )
    parser.add_argument(
        "-f",
        "--fresh",
        action="store_true",
        help="Start without loading previous user preferences",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=False,
        help="Dataset file to process (HDF5 format)",
        default="",
        metavar="FILE",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        type=str,
        required=False,
        help="Folder where to store the results",
        default="",
        metavar="FOLDER",
    )
    parser.add_argument(
        "-c",
        "--pyfai-config",
        type=str,
        default=None,
        help="PyFAI config file (.json)",
        metavar="FILE",
    )
    parser.add_argument(
        "--flat-dir",
        type=str,
        required=False,
        help=f"Folder containing flat-field files: flats.mat and old_flats.mat (default: {FLATFIELD_DEFAULT_DIR})",
        default=FLATFIELD_DEFAULT_DIR,
        metavar="FOLDER",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity level (-v: INFO, -vv:DEBUG)",
    )
    return parser


def main():
    logging.basicConfig(level=logging.WARNING)

    parser = create_argument_parser()
    args = parser.parse_args()

    logging.captureWarnings(True)

    if args.verbose == 0:
        logging.getLogger().setLevel(logging.ERROR)
        _logger.setLevel(logging.INFO)

    elif args.verbose == 1:
        logging.getLogger().setLevel(logging.WARNING)
        _logger.setLevel(logging.INFO)

    else:
        logging.getLogger().setLevel(logging.DEBUG)

    if sys.platform == "win32":
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
            "ESRF.id31pyfaiewoks"
        )

    app = qt.QApplication([])
    app.setOrganizationName("ESRF")
    app.setOrganizationDomain("esrf.fr")
    app.setApplicationName("id31pyfaiewoks")

    app.setWindowIcon(
        qt.QIcon(os.path.join(os.path.dirname(__file__), "integrate.svg"))
    )

    window = Id31FAIEwoksMainWindow()
    window.setAttribute(qt.Qt.WA_DeleteOnClose)

    window.setFlatFieldDirName(args.flat_dir)

    if not args.fresh:
        _logger.debug(
            "Launching application in default mode. Loading previous settings."
        )

        window.loadSettings()

    if args.input:
        raw_data = os.path.abspath(args.input)
        if os.path.isfile(raw_data) and raw_data.endswith(_FILE_EXTENSIONS):
            window.addRawDataFile(raw_data)
        else:
            _logger.error(f"Invalid raw data file path or format: {raw_data}")

    if args.output_dir:
        window._outputWidget.setOutputDirName(args.output_dir)

    if args.pyfai_config:
        config_file = os.path.abspath(args.pyfai_config)
        if os.path.isfile(config_file):
            window.setConfigFilePath(config_file)
        else:
            _logger.error(f"Invalid config file path: {config_file}")

    window.show()
    app.exec_()



from dataclasses import dataclass
import logging

import os
import h5py
import numpy
from silx.io.nxdata import NXdata

from ewoksxrpd.tasks.data_access import TaskWithDataAccess
from ewoksxrpd.tasks.utils.ascii_utils import ensure_parent_folder


logger = logging.getLogger(__name__)


@dataclass
class IntegratedPatterns:
    """Store multiple pyFAI integrated patterns"""

    radial: numpy.ndarray
    radial_name: str
    radial_units: str
    intensities: numpy.ndarray


class SaveNexusPatternsAsId31TomoHdf5(
    TaskWithDataAccess,
    input_names=[
        "scan_entry_url",
        "rot_name",
        "y_name",
        "nxdata_url",
        "output_filename",
    ],
    optional_input_names=["enabled", "overwrite"],
    output_names=["filename"],
):
    """Save integrated XRD patterns into an HDF5 file compatible with ID31 tomography workflow.

    Inputs:
    - scan_entry_url (str): HDF5 URL to the NXentry group containing the "measurement" group.
    - rot_name (str): Dataset name in "measurement" for rotation angles (e.g. "nth").
    - y_name (str): Dataset name in "measurement" for horizontal positions (e.g. "ny").
    - nxdata_url (str): HDF5 URL to the NXdata group containing integrated patterns.
    - output_filename (str): Path to the output HDF5 file.

    Optional inputs:
    - enabled (bool | Optional): True to enable saving, False to skip task (default: True).

    Output:
    - filename (str): Path to the written HDF5 file, or empty string if skipped.
    """

    def run(self):
        if not self.get_input_value("enabled", True):
            logger.info(
                f"Task {self.__class__.__qualname__} is disabled: No file is saved"
            )
            self.outputs.filename = ""
            return

        output_filename = self.inputs.output_filename
        ensure_parent_folder(output_filename)

        with self.open_h5item(self.inputs.scan_entry_url) as scan_entry:
            measurement = scan_entry["measurement"]

            th_angles_dataset = measurement[self.inputs.rot_name]
            th_angles = th_angles_dataset[()]
            th_angles_units = th_angles_dataset.attrs.get("units", None)

            y_positions_dataset = measurement[self.inputs.y_name]
            y_positions = y_positions_dataset[()]
            y_positions_units = y_positions_dataset.attrs.get("units", None)

        with self.open_h5item(self.inputs.nxdata_url) as group:
            patterns = _read_nexus_integrated_patterns(group)

        if self.get_input_value("overwrite", False):
            mode = "w"
        else:
            mode = "a"

        if os.path.exists(output_filename) and not self.get_input_value(
            "overwrite", False
        ):
            raise FileExistsError(
                f"{output_filename} already exists. Use overwrite=True to replace it."
            )

        with h5py.File(output_filename, mode) as h5f:
            h5f["XRD"] = patterns.intensities

            h5f[patterns.radial_name] = patterns.radial
            if patterns.radial_units:
                h5f[patterns.radial_name].attrs["units"] = patterns.radial_units

            h5f["y"] = y_positions
            if y_positions_units is not None:
                h5f["y"].attrs["units"] = y_positions_units

            h5f["th"] = th_angles
            if th_angles_units is not None:
                h5f["th"].attrs["units"] = th_angles_units

        self.outputs.filename = output_filename


def _read_nexus_integrated_patterns(group: h5py.Group) -> IntegratedPatterns:
    """Read integrated patterns from a HDF5 NXdata group.

    It reads from both single (1D signal) or multi (2D signal) NXdata.
    """
    nxdata = NXdata(group)
    if not nxdata.is_valid:
        raise RuntimeError(
            f"Cannot parse NXdata group: {group.file.filename}::{group.name}"
        )
    if not (nxdata.signal_is_1d or nxdata.signal_is_2d):
        raise RuntimeError(
            f"Signal is not a 1D or 2D dataset: {group.file.filename}::{group.name}"
        )

    if nxdata.axes[-1] is None:
        radial = numpy.arange(nxdata.signal.shape[-1])
        radial_name = "radial"
        radial_units = ""
    else:
        axis_dataset = nxdata.axes[-1]
        radial = axis_dataset[()]
        radial_name = nxdata.axes_dataset_names[-1]
        radial_units = axis_dataset.attrs.get("units", "")

    intensities = numpy.atleast_2d(nxdata.signal)

    return IntegratedPatterns(radial, radial_name, radial_units, intensities)



import os
import logging

from silx.gui import qt, icons

from ewoksid31.app.integrate.utils import FilenameCompleterLineEdit

_logger = logging.getLogger(__name__)


class OutputWidget(qt.QGroupBox):
    def __init__(self):
        super().__init__("Output")

        gridLayout = qt.QGridLayout()

        spacer = qt.QSpacerItem(
            20, 10, qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        gridLayout.addItem(spacer, 0, 1, 1, 1)

        self._outputLineEdit = FilenameCompleterLineEdit()
        self._outputLineEdit.setPlaceholderText("/path/to/output/directory")

        outputDirButton = qt.QPushButton("Select...")
        outputDirButton.setToolTip("Define a new output directory.")
        outputDirButton.setIcon(icons.getQIcon("folder"))
        outputDirButton.clicked.connect(self._outputDirButtonClicked)

        self._asciiExportCheckBox = qt.QCheckBox()
        self._asciiExportCheckBox.setToolTip(
            "Enable to export data in ASCII format (.xye)."
        )

        self._tomoEnabledCheckBox = qt.QCheckBox("")
        self._tomoEnabledCheckBox.setChecked(False)
        self._tomoEnabledCheckBox.setToolTip(
            "Enable to export data in tomo-compatible HDF5 files."
        )
        self._tomoEnabledCheckBox.toggled.connect(self.toggleTomoWidgets)

        self._rotNameLineEdit = qt.QLineEdit()
        self._rotNameLineEdit.setPlaceholderText("{rot_name}")
        self._rotNameLineEdit.setToolTip(
            "Specify the Rotation angle dataset (e.g., nth)."
        )

        self._yNameLineEdit = qt.QLineEdit()
        self._yNameLineEdit.setPlaceholderText("{y_name}")
        self._yNameLineEdit.setToolTip(
            "Specify the radial position dataset (e.g., ny)."
        )
        self.toggleTomoWidgets(self._tomoEnabledCheckBox.isChecked())

        gridLayout.addWidget(qt.QLabel("Directory:"), 0, 0, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(self._outputLineEdit, 0, 1, 1, 3)
        gridLayout.addWidget(outputDirButton, 0, 4, 1, 1, qt.Qt.AlignRight)

        gridLayout.addWidget(qt.QLabel("ASCII export:"), 1, 0, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(self._asciiExportCheckBox, 1, 1, 1, 1, qt.Qt.AlignLeft)

        gridLayout.addWidget(
            qt.QLabel("XRD-CT tomo export:"), 2, 0, 1, 1, qt.Qt.AlignLeft
        )
        gridLayout.addWidget(self._tomoEnabledCheckBox, 2, 1, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(qt.QLabel("- Rot name:"), 3, 0, 1, 1, qt.Qt.AlignRight)
        gridLayout.addWidget(self._rotNameLineEdit, 3, 1, 1, 1, qt.Qt.AlignLeft)
        gridLayout.addWidget(qt.QLabel("- Y name:"), 4, 0, 1, 1, qt.Qt.AlignRight)
        gridLayout.addWidget(self._yNameLineEdit, 4, 1, 1, 1, qt.Qt.AlignLeft)

        self.setLayout(gridLayout)

    def getOutputDirName(self) -> str:
        """
        Returns the current output directory path from the line edit.
        """
        return self._outputLineEdit.text().strip()

    def setOutputDirName(self, path: str) -> None:
        """
        Update the output directory path in the line edit.
        """
        self._outputLineEdit.setText(os.path.abspath(path))

    def _outputDirButtonClicked(self) -> None:
        """
        Choose an output directory using a dialog and set it in the line edit.
        """
        currentPath = self.getOutputDirName()
        if currentPath and os.path.exists(currentPath):
            pass
        else:
            currentPath = os.getcwd()
            _logger.info(f"Path not found. Falling back to: {currentPath}")

        newPath = qt.QFileDialog.getExistingDirectory(
            self, "Choose output directory", currentPath
        )
        if newPath:
            self.setOutputDirName(newPath)

    def isAsciiExportEnabled(self) -> bool:
        return self._asciiExportCheckBox.isChecked()

    def setAsciiExportEnabled(self, value: bool):
        self._asciiExportCheckBox.setChecked(value)

    def toggleTomoWidgets(self, enabled):
        self._rotNameLineEdit.setEnabled(enabled)
        self._yNameLineEdit.setEnabled(enabled)

    def isTomoEnabled(self) -> bool:
        return self._tomoEnabledCheckBox.isChecked()

    def setTomoEnabled(self, value: bool):
        self._tomoEnabledCheckBox.setChecked(value)

    def getTomoRotName(self) -> str:
        return self._rotNameLineEdit.text().strip()

    def setTomoRotName(self, value: str):
        self._rotNameLineEdit.setText(value)

    def getTomoYName(self) -> str:
        return self._yNameLineEdit.text().strip()

    def setTomoYName(self, value: str):
        self._yNameLineEdit.setText(value)



"""
Usage:

python streamline_autocalib.py -v -o output_folder \
    --calibrant LaB6_SRM660c --max_rings 35 \
    --calibration /data/visitor/in1176/id31/20241022/RAW_DATA/ESRF_HR_01/ESRF_HR_01_0003/ESRF_HR_01_0003.h5 \
    --flat-dir /data/visitor/in1176/id31/20241022/PROCESSED_DATA/extra_files/ \
    --pyfai-config /data/visitor/in1176/id31/20241022/PROCESSED_DATA/calibration/PDF.json

"""

import argparse
import logging
import os

import h5py
from ewoks import execute_graph
from ewoksutils.task_utils import task_inputs

try:
    from ewoksjob.client import submit
except ImportError:
    submit = None

from .utils import (
    FLATFIELD_DEFAULT_DIR,
    NEWFLAT_FILENAME,
    OLDFLAT_FILENAME,
    print_inputs,
)


def generate_inputs(
    workflow,
    output_dir,
    pyfai_config,
    integration_options,
    detector_name,
    energy,
    newflat,
    oldflat,
    bliss_scan_url,
    integrate_image_url,
    calibrant,
    calib_ring_detector_name,
    calibrate_image_url,
    max_rings,
):
    inputs = [
        {
            "task_identifier": "PyFaiConfig",
            "name": "filename",
            "value": pyfai_config,
        },
        {
            "task_identifier": "PyFaiConfig",
            "name": "integration_options",
            "value": integration_options,
        },
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "newflat",
            "value": newflat,
        },
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "oldflat",
            "value": oldflat,
        },
        {
            "task_identifier": "FlatFieldFromEnergy",
            "name": "energy",
            "value": energy,
        },
        {
            "task_identifier": "Integrate1D",
            "name": "image",
            "value": integrate_image_url,
        },
        {
            "task_identifier": "SaveNexusPattern1D",
            "name": "url",
            "value": os.path.join(output_dir, "result.h5"),
        },
        {
            "task_identifier": "SaveNexusPattern1D",
            "name": "bliss_scan_url",
            "value": bliss_scan_url,
        },
        {
            "task_identifier": "DiagnoseIntegrate1D",
            "name": "filename",
            "value": os.path.join(output_dir, "integrate.png"),
        },
    ]

    if calibrant:
        inputs += [
            {
                "task_identifier": "CalibrateSingle",
                "name": "image",
                "value": calibrate_image_url,
            },
            {
                "task_identifier": "CalibrateSingle",
                "name": "fixed",
                "value": ["energy"],
            },
            {
                "task_identifier": "CalibrateSingle",
                "name": "robust",
                "value": False,
            },
            {
                "task_identifier": "CalibrateSingle",
                "name": "ring_detector",
                "value": calib_ring_detector_name,
            },
            {
                "task_identifier": "CalibrateSingle",
                "name": "max_rings",
                "value": max_rings,
            },
            {
                "task_identifier": "DiagnoseCalibrateSingleResults",
                "name": "image",
                "value": calibrate_image_url,
            },
            {
                "task_identifier": "DiagnoseCalibrateSingleResults",
                "name": "filename",
                "value": os.path.join(output_dir, "ring_detection.png"),
            },
            {
                "task_identifier": "PyFaiConfig",
                "name": "calibrant",
                "value": calibrant,
            },
            {
                "task_identifier": "DiagnoseIntegrate1D",
                "name": "calibrant",
                "value": calibrant,
            },
            {
                "task_identifier": "DiagnoseIntegrate1D",
                "label": "diagnose_calibrate",
                "name": "filename",
                "value": os.path.join(output_dir, "calibrate.png"),
            },
        ]

    for unit in ("q", "2th", "q_no_sigmaclip"):
        inputs += task_inputs(
            task_identifier="SaveNexusPattern1D",
            label=f"save_{unit}_hdf5",
            inputs={
                "nxprocess_name": f"{detector_name}_integrate_{unit}",
                "nxmeasurement_name": f"{detector_name}_integrated_{unit}",
                "metadata": {
                    f"{detector_name}_integrate_{unit}": {
                        "configuration": {"workflow": workflow}
                    }
                },
            },
        )

    for unit in ("q", "2th"):
        inputs += [
            {
                "task_identifier": "SaveAsciiPattern1D",
                "label": f"save_{unit}_ascii",
                "name": "filename",
                "value": os.path.join(output_dir, f"{unit}_result.xye"),
            },
        ]

    return inputs


def create_argument_parser():
    parser = argparse.ArgumentParser(description="")

    parser.add_argument(
        "-o",
        "--output-dir",
        type=str,
        required=True,
        help="Folder where to store the results",
        metavar="FOLDER",
    )
    parser.add_argument(
        "--worker",
        action="store_true",
        help="Execute workflows on ewoks worker instead of current environment",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity",
    )

    calibration_group = parser.add_argument_group("Calibration")
    calibration_group.add_argument(
        "--calibrant",
        type=str,
        required=True,
        help="Name of the calibrant used for calibration image",
    )
    calibration_group.add_argument(
        "--calibration",
        type=str,
        required=True,
        dest="calibration_filename",
        help="Dataset file to use for auto-calibration",
        metavar="FILE",
    )
    calibration_group.add_argument(
        "--max_rings",
        type=int,
        required=False,
        nargs="+",
        help="Number of rings to use. Use multiple values to refine over an increasing number of rings",
        default=[35],
    )

    integration_group = parser.add_argument_group("Integration")
    integration_group.add_argument(
        "-c",
        "--pyfai-config",
        type=str,
        required=True,
        help="PyFAI config file (.json)",
        metavar="FILE",
    )
    integration_group.add_argument(
        "-i",
        "--input",
        type=str,
        required=False,
        help="Dataset file to process (HDF5 format)",
        default="",
        metavar="FILE",
    )
    integration_group.add_argument(
        "--flat-dir",
        type=str,
        required=False,
        help="Folder containing flat field files: flats.mat and old_flats.mat",
        default=FLATFIELD_DEFAULT_DIR,
        metavar="FOLDER",
    )

    return parser


def main():
    logging.basicConfig(level=logging.WARNING)

    parser = create_argument_parser()
    options = parser.parse_args()

    if options.verbose != 0:
        logging.getLogger().setLevel(
            logging.INFO if options.verbose == 1 else logging.DEBUG
        )

    workflow = "streamline_with_calib_with_flat.json"
    workflow_load_options = {"root_module": "ewoksid31.workflows"}

    integration_options = {"nbpt_rad": 3000}
    detector_name = "p3"
    calib_ring_detector_name = "PilatusCdTe2M"

    input_filename = options.input if options.input else options.calibration_filename
    bliss_scan_url = f"{input_filename}::/1.1"
    integrate_image_url = (
        f"silx://{input_filename}?path=/1.1/measurement/{detector_name}&slice=0"
    )

    calibrate_image_url = f"silx://{options.calibration_filename}?path=/1.1/measurement/{detector_name}&slice=0"

    # ewoksxrpd CalibrateSingle compatibilty: list not supported
    max_rings = (
        options.max_rings[0] if len(options.max_rings) == 1 else options.max_rings
    )

    with h5py.File(input_filename, "r") as h5f:
        energy = h5f["/1.1/instrument/positioners/energy"][()]

    inputs = generate_inputs(
        workflow,
        os.path.abspath(options.output_dir),
        options.pyfai_config,
        integration_options,
        detector_name,
        energy,
        os.path.join(options.flat_dir, NEWFLAT_FILENAME),
        os.path.join(options.flat_dir, OLDFLAT_FILENAME),
        bliss_scan_url,
        integrate_image_url,
        options.calibrant,
        calib_ring_detector_name,
        calibrate_image_url,
        max_rings,
    )
    print("Worflow inputs:")
    print_inputs(inputs)

    if os.path.exists(options.output_dir):
        raise RuntimeError(f"Output directory already exists: {options.output_dir}")

    os.makedirs(options.output_dir)

    print("Execute workflow...")
    convert_destination = os.path.join(options.output_dir, "workflow.json")
    if options.worker:
        if submit is None:
            raise RuntimeError("Cannot submit workflow to worker: Install ewoksjob")
        _ = submit(
            args=(workflow,),
            kwargs={
                "inputs": inputs,
                "convert_destination": convert_destination,
                "load_options": workflow_load_options,
            },
        )
    else:
        _ = execute_graph(
            workflow,
            inputs=inputs,
            convert_destination=convert_destination,
            load_options=workflow_load_options,
        )
    print("Done")


if __name__ == "__main__":
    import sys

    sys.exit(main())



import json
from dataclasses import dataclass
from pathlib import Path

import numpy
import pytest
import scipy.io

from ..app.integrate.id31_pyfai_ewoks import Id31FAIEwoksMainWindow
from .conftest import DETECTOR_SHAPE


@dataclass
class DummyH5Node:
    physical_filename: str
    physical_name: str


PONI = {
    "application": "pyfai-integrate",
    "version": 3,
    "wavelength": 1.6531226457760035e-11,
    "dist": 1.014612139238891,
    "poni1": 0.21340662994315895,
    "poni2": 0.13912764384981186,
    "rot1": -0.004822100080733148,
    "rot2": 0.0005542810978441514,
    "rot3": 0.0,
    "detector": "Pilatus_CdTe_2M",
}


def _create_flats(flat_dir: Path, n=5):
    energy = numpy.linspace(0, 10, n)
    flat = numpy.ones((*DETECTOR_SHAPE, n))
    scipy.io.savemat(flat_dir / "flats.mat", mdict={"E": energy, "F": flat})
    scipy.io.savemat(flat_dir / "flats_old.mat", mdict={"Eold": energy, "Fold": flat})


@pytest.mark.parametrize("overwrite", (True, False))
def test_integrate(qtbot, tmp_path, tomo_p3_scan, overwrite):
    window = Id31FAIEwoksMainWindow()
    qtbot.addWidget(window)

    with open(tmp_path / "config.json", "w") as cfg:
        json.dump(PONI, cfg)
    window.setConfigFilePath(str(tmp_path / "config.json"))

    _create_flats(tmp_path)
    window.setFlatFieldDirName(tmp_path)

    window._2thCheckBox.click()
    window._outputWidget.setTomoEnabled(False)

    output_dir = tmp_path / "PROCESSED_DATA"
    output_dir.mkdir()
    window._outputWidget.setOutputDirName(output_dir)

    integration_output_file = output_dir / f"{tomo_p3_scan.stem}.h5"

    # Launch processing manually (could be done via the GUI later)
    scan = DummyH5Node(physical_filename=str(tomo_p3_scan), physical_name="/2.1")
    inputParameters = window._prepareScans([scan])

    window._executeWorkflowForParams(inputParameters)
    inputParameters[0]["inputs"].append(
        {"name": "overwrite", "value": overwrite, "all": True}
    )

    # Try to overwrite the files just created
    if overwrite:
        window._executeWorkflowForParams(inputParameters)
    else:
        with pytest.raises(RuntimeError) as exc:
            window._executeWorkflowForParams(inputParameters)
        original_exc = exc.value.__cause__
        assert isinstance(original_exc, ValueError)
        assert (
            str(original_exc)
            == "Unable to synchronously create group (name already exists)"
        )

    assert integration_output_file.exists()
